@@ -28,7 +28,7 @@
 #define CONFIG_H
 
 #define PACKAGE			"dhcpcd"
-#define VERSION			"6.8.2"
+#define VERSION			"6.9.2"
 
 #ifndef CONFIG
 # define CONFIG			SYSCONFDIR "/" PACKAGE ".conf"
@@ -46,7 +46,7 @@
 # define SECRET			SYSCONFDIR "/" PACKAGE ".secret"
 #endif
 #ifndef LEASEFILE
-# define LEASEFILE		DBDIR "/" PACKAGE "-%s%s%s.lease"
+# define LEASEFILE		DBDIR "/" PACKAGE "-%s%s.lease"
 #endif
 #ifndef LEASEFILE6
 # define LEASEFILE6		LEASEFILE "6"
