@@ -27,6 +27,8 @@
 
 #include <sys/utsname.h>
 
+#include <arpa/nameser.h>
+
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -43,6 +45,13 @@
 #include "if.h"
 #include "ipv6.h"
 
+/* Support very old arpa/nameser.h as found in OpenBSD */
+#ifndef NS_MAXDNAME
+#define NS_MAXCDNAME MAXCDNAME
+#define NS_MAXDNAME MAXDNAME
+#define NS_MAXLABEL MAXLABEL
+#endif
+
 void
 dhcp_print_option_encoding(const struct dhcp_opt *opt, int cols)
 {
@@ -76,7 +85,9 @@ dhcp_print_option_encoding(const struct dhcp_opt *opt, int cols)
 		printf(" ip6address");
 	else if (opt->type & FLAG)
 		printf(" flag");
-	else if (opt->type & RFC3397)
+	else if (opt->type & BITFLAG)
+		printf(" bitflags");
+	else if (opt->type & RFC1035)
 		printf(" domain");
 	else if (opt->type & DOMAIN)
 		printf(" dname");
@@ -86,9 +97,9 @@ dhcp_print_option_encoding(const struct dhcp_opt *opt, int cols)
 		printf(" raw");
 	else if (opt->type & BINHEX)
 		printf(" binhex");
-	else if (opt->type & STRING)	
+	else if (opt->type & STRING)
 		printf(" string");
-	if (opt->type & RFC3361)	
+	if (opt->type & RFC3361)
 		printf(" rfc3361");
 	if (opt->type & RFC3442)
 		printf(" rfc3442");
@@ -262,12 +273,12 @@ encode_rfc1035(const char *src, uint8_t *dst)
 	return len;
 }
 
-/* Decode an RFC3397 DNS search order option into a space
+/* Decode an RFC1035 DNS search order option into a space
  * separated string. Returns length of string (including
  * terminating zero) or zero on error. out may be NULL
  * to just determine output length. */
 ssize_t
-decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
+decode_rfc1035(char *out, size_t len, const uint8_t *p, size_t pl)
 {
 	const char *start;
 	size_t start_len, l, count;
@@ -275,6 +286,11 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 	int hops;
 	uint8_t ltype;
 
+	if (pl > NS_MAXCDNAME) {
+		errno = E2BIG;
+		return -1;
+	}
+
 	count = 0;
 	start = out;
 	start_len = len;
@@ -287,8 +303,13 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 		 * the name isn't fully qualified (ie, not terminated) */
 		while (q < e && (l = (size_t)*q++)) {
 			ltype = l & 0xc0;
-			if (ltype == 0x80 || ltype == 0x40)
+			if (ltype == 0x80 || ltype == 0x40) {
+				/* Currently reserved for future use as noted
+				 * in RFC1035 4.1.4 as the 10 and 01
+				 * combinations. */
+				errno = ENOTSUP;
 				return -1;
+			}
 			else if (ltype == 0xc0) { /* pointer */
 				if (q == e) {
 					errno = ERANGE;
@@ -321,6 +342,10 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 						errno = ENOBUFS;
 						return -1;
 					}
+					if (l + 1 > NS_MAXLABEL) {
+						errno = EINVAL;
+						return -1;
+					}
 					memcpy(out, q, l);
 					out += l;
 					*out++ = '.';
@@ -348,6 +373,10 @@ decode_rfc3397(char *out, size_t len, const uint8_t *p, size_t pl)
 	if (count)
 		/* Don't count the trailing NUL */
 		count--;
+	if (count > NS_MAXDNAME) {
+		errno = E2BIG;
+		return -1;
+	}
 	return (ssize_t)count;
 }
 
@@ -397,7 +426,7 @@ valid_domainname(char *lbl, int type)
 		    !start && *lbl != ' ' && *lbl != '\0') ||
 		    isalnum(c))
 		{
-			if (++len > 63) {
+			if (++len > NS_MAXLABEL) {
 				errno = ERANGE;
 				errset = 1;
 				break;
@@ -524,60 +553,51 @@ print_string(char *dst, size_t len, int type, const uint8_t *data, size_t dl)
 
 #define ADDRSZ		4
 #define ADDR6SZ		16
-static size_t
+static ssize_t
 dhcp_optlen(const struct dhcp_opt *opt, size_t dl)
 {
 	size_t sz;
 
-	if (dl == 0)
-		return 0;
-
 	if (opt->type == 0 ||
 	    opt->type & (STRING | BINHEX | RFC3442 | RFC5969))
 	{
 		if (opt->len) {
 			if ((size_t)opt->len > dl)
-				return 0;
-			return (size_t)opt->len;
+				return -1;
+			return (ssize_t)opt->len;
 		}
-		return dl;
+		return (ssize_t)dl;
 	}
 
 	if ((opt->type & (ADDRIPV4 | ARRAY)) == (ADDRIPV4 | ARRAY)) {
 		if (dl < ADDRSZ)
-			return 0;
-		return dl - (dl % ADDRSZ);
+			return -1;
+		return (ssize_t)(dl - (dl % ADDRSZ));
 	}
 
 	if ((opt->type & (ADDRIPV6 | ARRAY)) == (ADDRIPV6 | ARRAY)) {
 		if (dl < ADDR6SZ)
-			return 0;
-		return dl - (dl % ADDR6SZ);
+			return -1;
+		return (ssize_t)(dl - (dl % ADDR6SZ));
 	}
 
 	if (opt->type & (UINT32 | ADDRIPV4))
 		sz = sizeof(uint32_t);
 	else if (opt->type & UINT16)
 		sz = sizeof(uint16_t);
-	else if (opt->type & UINT8)
+	else if (opt->type & (UINT8 | BITFLAG))
 		sz = sizeof(uint8_t);
 	else if (opt->type & ADDRIPV6)
 		sz = ADDR6SZ;
 	else
 		/* If we don't know the size, assume it's valid */
-		return dl;
-	return (dl < sz ? 0 : sz);
+		return (ssize_t)dl;
+	return dl < sz ? -1 : (ssize_t)sz;
 }
 
-#ifdef INET6
-#define PO_IFNAME
-#else
-#define PO_IFNAME __unused
-#endif
-
-ssize_t
-print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
-    PO_IFNAME const char *ifname)
+static ssize_t
+print_option(char *s, size_t len, const struct dhcp_opt *opt,
+    const uint8_t *data, size_t dl, const char *ifname)
 {
 	const uint8_t *e, *t;
 	uint16_t u16;
@@ -589,41 +609,45 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 	size_t l;
 	char *tmp;
 
-	if (type & RFC3397) {
-		sl = decode_rfc3397(NULL, 0, data, dl);
+#ifndef INET6
+	UNUSED(ifname);
+#endif
+
+	if (opt->type & RFC1035) {
+		sl = decode_rfc1035(NULL, 0, data, dl);
 		if (sl == 0 || sl == -1)
 			return sl;
 		l = (size_t)sl + 1;
 		tmp = malloc(l);
 		if (tmp == NULL)
 			return -1;
-		decode_rfc3397(tmp, l, data, dl);
-		sl = print_string(s, len, type, (uint8_t *)tmp, l - 1);
+		decode_rfc1035(tmp, l, data, dl);
+		sl = print_string(s, len, opt->type, (uint8_t *)tmp, l - 1);
 		free(tmp);
 		return sl;
 	}
 
 #ifdef INET
-	if (type & RFC3361) {
+	if (opt->type & RFC3361) {
 		if ((tmp = decode_rfc3361(data, dl)) == NULL)
 			return -1;
 		l = strlen(tmp);
-		sl = print_string(s, len, type, (uint8_t *)tmp, l);
+		sl = print_string(s, len, opt->type, (uint8_t *)tmp, l);
 		free(tmp);
 		return sl;
 	}
 
-	if (type & RFC3442)
+	if (opt->type & RFC3442)
 		return decode_rfc3442(s, len, data, dl);
 
-	if (type & RFC5969)
+	if (opt->type & RFC5969)
 		return decode_rfc5969(s, len, data, dl);
 #endif
 
-	if (type & STRING)
-		return print_string(s, len, type, data, dl);
+	if (opt->type & STRING)
+		return print_string(s, len, opt->type, data, dl);
 
-	if (type & FLAG) {
+	if (opt->type & FLAG) {
 		if (s) {
 			*s++ = '1';
 			*s = '\0';
@@ -631,27 +655,50 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 		return 1;
 	}
 
+	if (opt->type & BITFLAG) {
+		/* bitflags are a string, MSB first, such as ABCDEFGH
+		 * where A is 10000000, B is 01000000, etc. */
+		bytes = 0;
+		for (l = 0, sl = sizeof(opt->bitflags) - 1;
+		    l < sizeof(opt->bitflags);
+		    l++, sl--)
+		{
+			/* Don't print NULL or 0 flags */
+			if (opt->bitflags[l] != '\0' &&
+			    opt->bitflags[l] != '0' &&
+			    *data & (1 << sl))
+			{
+				if (s)
+					*s++ = opt->bitflags[l];
+				bytes++;
+			}
+		}
+		if (s)
+			*s = '\0';
+		return bytes;
+	}
+
 	if (!s) {
-		if (type & UINT8)
+		if (opt->type & UINT8)
 			l = 3;
-		else if (type & UINT16) {
+		else if (opt->type & UINT16) {
 			l = 5;
 			dl /= 2;
-		} else if (type & SINT16) {
+		} else if (opt->type & SINT16) {
 			l = 6;
 			dl /= 2;
-		} else if (type & UINT32) {
+		} else if (opt->type & UINT32) {
 			l = 10;
 			dl /= 4;
-		} else if (type & SINT32) {
+		} else if (opt->type & SINT32) {
 			l = 11;
 			dl /= 4;
-		} else if (type & ADDRIPV4) {
+		} else if (opt->type & ADDRIPV4) {
 			l = 16;
 			dl /= 4;
 		}
 #ifdef INET6
-		else if (type & ADDRIPV6) {
+		else if (opt->type & ADDRIPV6) {
 			e = data + dl;
 			l = 0;
 			while (data < e) {
@@ -680,36 +727,36 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 			bytes++;
 			len--;
 		}
-		if (type & UINT8) {
+		if (opt->type & UINT8) {
 			sl = snprintf(s, len, "%u", *data);
 			data++;
-		} else if (type & UINT16) {
+		} else if (opt->type & UINT16) {
 			memcpy(&u16, data, sizeof(u16));
 			u16 = ntohs(u16);
 			sl = snprintf(s, len, "%u", u16);
 			data += sizeof(u16);
-		} else if (type & SINT16) {
+		} else if (opt->type & SINT16) {
 			memcpy(&u16, data, sizeof(u16));
 			s16 = (int16_t)ntohs(u16);
 			sl = snprintf(s, len, "%d", s16);
 			data += sizeof(u16);
-		} else if (type & UINT32) {
+		} else if (opt->type & UINT32) {
 			memcpy(&u32, data, sizeof(u32));
 			u32 = ntohl(u32);
 			sl = snprintf(s, len, "%u", u32);
 			data += sizeof(u32);
-		} else if (type & SINT32) {
+		} else if (opt->type & SINT32) {
 			memcpy(&u32, data, sizeof(u32));
 			s32 = (int32_t)ntohl(u32);
 			sl = snprintf(s, len, "%d", s32);
 			data += sizeof(u32);
-		} else if (type & ADDRIPV4) {
+		} else if (opt->type & ADDRIPV4) {
 			memcpy(&addr.s_addr, data, sizeof(addr.s_addr));
 			sl = snprintf(s, len, "%s", inet_ntoa(addr));
 			data += sizeof(addr.s_addr);
 		}
 #ifdef INET6
-		else if (type & ADDRIPV6) {
+		else if (opt->type & ADDRIPV6) {
 			ssize_t r;
 
 			r = ipv6_printaddr(s, len, data, ifname);
@@ -732,7 +779,7 @@ print_option(char *s, size_t len, int type, const uint8_t *data, size_t dl,
 
 int
 dhcp_set_leasefile(char *leasefile, size_t len, int family,
-    const struct interface *ifp, const char *extra)
+    const struct interface *ifp)
 {
 	char ssid[len];
 
@@ -759,7 +806,7 @@ dhcp_set_leasefile(char *leasefile, size_t len, int family,
 		ssid[0] = '\0';
 	return snprintf(leasefile, len,
 	    family == AF_INET ? LEASEFILE : LEASEFILE6,
-	    ifp->name, ssid, extra);
+	    ifp->name, ssid);
 }
 
 static size_t
@@ -773,7 +820,7 @@ dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 
 	if (opt->len && opt->len < ol)
 		ol = opt->len;
-	len = print_option(NULL, 0, opt->type, od, ol, ifname);
+	len = print_option(NULL, 0, opt, od, ol, ifname);
 	if (len < 0)
 		return 0;
 	if (vname)
@@ -795,7 +842,7 @@ dhcp_envoption1(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 	else
 		v += snprintf(val, e, "%s=", prefix);
 	if (len != 0)
-		print_option(v, (size_t)len + 1, opt->type, od, ol, ifname);
+		print_option(v, (size_t)len + 1, opt, od, ol, ifname);
 	return e;
 }
 
@@ -808,6 +855,7 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
     const uint8_t *od, size_t ol)
 {
 	size_t e, i, n, eos, eol;
+	ssize_t eo;
 	unsigned int eoc;
 	const uint8_t *eod;
 	int ov;
@@ -816,7 +864,8 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 
 	/* If no embedded or encapsulated options, it's easy */
 	if (opt->embopts_len == 0 && opt->encopts_len == 0) {
-		if (dhcp_envoption1(ctx, env == NULL ? NULL : &env[0],
+		if (!(opt->type & RESERVED) &&
+		    dhcp_envoption1(ctx, env == NULL ? NULL : &env[0],
 		    prefix, opt, 1, od, ol, ifname))
 			return 1;
 		return 0;
@@ -850,19 +899,42 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 	 * is a fixed layout */
 	n = 0;
 	for (i = 0, eopt = opt->embopts; i < opt->embopts_len; i++, eopt++) {
-		e = dhcp_optlen(eopt, ol);
-		if (e == 0)
-			/* Report error? */
-			return 0;
+		eo = dhcp_optlen(eopt, ol);
+		if (eo == -1) {
+			if (env == NULL)
+				logger(ctx, LOG_ERR,
+				    "%s: %s: malformed embedded option %d:%d",
+				    ifname, __func__, opt->option,
+				    eopt->option);
+			goto out;
+		}
+		if (eo == 0) {
+			/* An option was expected, but there is no data
+			 * data for it.
+			 * This may not be an error as some options like
+			 * DHCP FQDN in RFC4702 have a string as the last
+			 * option which is optional.
+			 * FIXME: Add an flag to the options to indicate
+			 * wether this is allowable or not. */
+			if (env == NULL &&
+			    (ol != 0 || i + 1 < opt->embopts_len))
+				logger(ctx, LOG_ERR,
+				    "%s: %s: malformed embedded option %d:%d",
+				    ifname, __func__, opt->option,
+				    eopt->option);
+			goto out;
+		}
 		/* Use the option prefix if the embedded option
 		 * name is different.
 		 * This avoids new_fqdn_fqdn which would be silly. */
-		ov = strcmp(opt->var, eopt->var);
-		if (dhcp_envoption1(ctx, env == NULL ? NULL : &env[n],
-		    pfx, eopt, ov, od, e, ifname))
-			n++;
-		od += e;
-		ol -= e;
+		if (!(eopt->type & RESERVED)) {
+			ov = strcmp(opt->var, eopt->var);
+			if (dhcp_envoption1(ctx, env == NULL ? NULL : &env[n],
+			    pfx, eopt, ov, od, (size_t)eo, ifname))
+				n++;
+		}
+		od += (size_t)eo;
+		ol -= (size_t)eo;
 	}
 
 	/* Enumerate our encapsulated options */
@@ -906,6 +978,7 @@ dhcp_envoption(struct dhcpcd_ctx *ctx, char **env, const char *prefix,
 		}
 	}
 
+out:
 	if (env)
 		free(pfx);
 
