@@ -168,30 +168,9 @@ get_option(struct dhcpcd_ctx *ctx,
 
 	while (p < e) {
 		o = *p++;
-		if (o == opt) {
-			if (op) {
-				if (!ctx->opt_buffer) {
-					ctx->opt_buffer =
-					    malloc(DHCP_OPTION_LEN +
-					    BOOTFILE_LEN + SERVERNAME_LEN);
-					if (ctx->opt_buffer == NULL)
-						return NULL;
-				}
-				if (!bp)
-					bp = ctx->opt_buffer;
-				memcpy(bp, op, ol);
-				bp += ol;
-			}
-			ol = *p;
-			if (p + ol > e) {
-				errno = EINVAL;
-				return NULL;
-			}
-			op = p + 1;
-			bl += ol;
-		}
 		switch (o) {
 		case DHO_PAD:
+			/* No length to read */
 			continue;
 		case DHO_END:
 			if (overl & 1) {
@@ -206,17 +185,48 @@ get_option(struct dhcpcd_ctx *ctx,
 				e = p + sizeof(dhcp->servername);
 			} else
 				goto exit;
-			break;
-		case DHO_OPTIONSOVERLOADED:
+			/* No length to read */
+			continue;
+		}
+
+		/* Check we can read the length */
+		if (p == e) {
+			errno = EINVAL;
+			return NULL;
+		}
+		l = *p++;
+
+		if (o == DHO_OPTIONSOVERLOADED) {
 			/* Ensure we only get this option once by setting
 			 * the last bit as well as the value.
 			 * This is valid because only the first two bits
 			 * actually mean anything in RFC2132 Section 9.3 */
-			if (!overl)
-				overl = 0x80 | p[1];
-			break;
+			if (l == 1 && !overl)
+				overl = 0x80 | p[0];
+		}
+
+		if (o == opt) {
+			if (op) {
+				if (!ctx->opt_buffer) {
+					ctx->opt_buffer =
+					    malloc(DHCP_OPTION_LEN +
+					    BOOTFILE_LEN + SERVERNAME_LEN);
+					if (ctx->opt_buffer == NULL)
+						return NULL;
+				}
+				if (!bp)
+					bp = ctx->opt_buffer;
+				memcpy(bp, op, ol);
+				bp += ol;
+			}
+			ol = l;
+			if (p + ol >= e) {
+				errno = EINVAL;
+				return NULL;
+			}
+			op = p;
+			bl += ol;
 		}
-		l = *p++;
 		p += l;
 	}
 
@@ -265,6 +275,23 @@ get_option_uint32(struct dhcpcd_ctx *ctx,
 	return 0;
 }
 
+static int
+get_option_uint16(struct dhcpcd_ctx *ctx,
+    uint16_t *i, const struct dhcp_message *dhcp, uint8_t option)
+{
+	const uint8_t *p;
+	size_t len;
+	uint16_t d;
+
+	p = get_option(ctx, dhcp, option, &len);
+	if (!p || len < (ssize_t)sizeof(d))
+		return -1;
+	memcpy(&d, p, sizeof(d));
+	if (i)
+		*i = ntohs(d);
+	return 0;
+}
+
 static int
 get_option_uint8(struct dhcpcd_ctx *ctx,
     uint8_t *i, const struct dhcp_message *dhcp, uint8_t option)
@@ -418,12 +445,12 @@ decode_rfc3361(const uint8_t *data, size_t dl)
 	dl--;
 	switch (enc) {
 	case 0:
-		if ((r = decode_rfc3397(NULL, 0, data, dl)) > 0) {
+		if ((r = decode_rfc1035(NULL, 0, data, dl)) > 0) {
 			l = (size_t)r;
 			sip = malloc(l);
 			if (sip == NULL)
 				return 0;
-			decode_rfc3397(sip, l, data, dl);
+			decode_rfc1035(sip, l, data, dl);
 		}
 		break;
 	case 1:
@@ -573,7 +600,7 @@ route_netmask(uint32_t ip_in)
 /* We need to obey routing options.
  * If we have a CSR then we only use that.
  * Otherwise we add static routes and then routers. */
-struct rt_head *
+static struct rt_head *
 get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 {
 	struct if_options *ifo = ifp->options;
@@ -644,6 +671,14 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 				free(route);
 				continue;
 			}
+			/* A host route is normally set by having the
+			 * gateway match the destination or assigned address */
+			if (route->gate.s_addr == route->dest.s_addr ||
+			    route->gate.s_addr == dhcp->yiaddr)
+			{
+				route->gate.s_addr = htonl(INADDR_ANY);
+				route->net.s_addr = htonl(INADDR_BROADCAST);
+			}
 			route->net.s_addr = route_netmask(route->dest.s_addr);
 			TAILQ_INSERT_TAIL(routes, route, next);
 		}
@@ -671,6 +706,40 @@ get_option_routes(struct interface *ifp, const struct dhcp_message *dhcp)
 	return routes;
 }
 
+uint16_t
+dhcp_get_mtu(const struct interface *ifp)
+{
+	const struct dhcp_message *dhcp;
+	uint16_t mtu;
+
+	if ((dhcp = D_CSTATE(ifp)->new) == NULL ||
+	    has_option_mask(ifp->options->nomask, DHO_MTU) ||
+	    get_option_uint16(ifp->ctx, &mtu, dhcp, DHO_MTU) == -1)
+		return 0;
+	return mtu;
+}
+
+/* Grab our routers from the DHCP message and apply any MTU value
+ * the message contains */
+struct rt_head *
+dhcp_get_routes(struct interface *ifp)
+{
+	struct rt_head *routes;
+	uint16_t mtu;
+	const struct dhcp_message *dhcp;
+
+	dhcp = D_CSTATE(ifp)->new;
+	routes = get_option_routes(ifp, dhcp);
+	if ((mtu = dhcp_get_mtu(ifp)) != 0) {
+		struct rt *rt;
+
+		TAILQ_FOREACH(rt, routes, next) {
+			rt->mtu = mtu;
+		}
+	}
+	return routes;
+}
+
 #define PUTADDR(_type, _val)						      \
 	{								      \
 		*p++ = _type;						      \
@@ -718,7 +787,6 @@ make_message(struct dhcp_message **message,
 	struct if_options *ifo = ifp->options;
 	const struct dhcp_state *state = D_CSTATE(ifp);
 	const struct dhcp_lease *lease = &state->lease;
-	time_t up = uptime() - state->start_uptime;
 	char hbuf[HOSTNAME_MAX_LEN + 1];
 	const char *hostname;
 	const struct vivco *vivco;
@@ -758,10 +826,14 @@ make_message(struct dhcp_message **message,
 		dhcp->flags = htons(BROADCAST_FLAG);
 
 	if (type != DHCP_DECLINE && type != DHCP_RELEASE) {
-		if (up < 0 || up > (time_t)UINT16_MAX)
+		struct timespec tv;
+
+		clock_gettime(CLOCK_MONOTONIC, &tv);
+		timespecsub(&tv, &state->started, &tv);
+		if (tv.tv_sec < 0 || tv.tv_sec > (time_t)UINT16_MAX)
 			dhcp->secs = htons((uint16_t)UINT16_MAX);
 		else
-			dhcp->secs = htons((uint16_t)up);
+			dhcp->secs = htons((uint16_t)tv.tv_sec);
 	}
 	dhcp->xid = htonl(state->xid);
 	dhcp->cookie = htonl(MAGIC_COOKIE);
@@ -828,21 +900,27 @@ make_message(struct dhcp_message **message,
 		if (!(ifo->options & DHCPCD_BOOTP)) {
 			int mtu;
 
-			*p++ = DHO_MAXMESSAGESIZE;
-			*p++ = 2;
-			mtu = if_getmtu(ifp->name);
-			if (mtu < MTU_MIN) {
-				if (if_setmtu(ifp->name, MTU_MIN) == 0)
-					sz = MTU_MIN;
+			if ((mtu = if_getmtu(ifp)) == -1)
+				logger(ifp->ctx, LOG_ERR,
+				    "%s: if_getmtu: %m", ifp->name);
+			else if (mtu < MTU_MIN) {
+				if (if_setmtu(ifp, MTU_MIN) == -1)
+					logger(ifp->ctx, LOG_ERR,
+					    "%s: if_setmtu: %m", ifp->name);
+				mtu = MTU_MIN;
 			} else if (mtu > MTU_MAX) {
 				/* Even though our MTU could be greater than
 				 * MTU_MAX (1500) dhcpcd does not presently
 				 * handle DHCP packets any bigger. */
 				mtu = MTU_MAX;
 			}
-			sz = htons((uint16_t)mtu);
-			memcpy(p, &sz, 2);
-			p += 2;
+			if (mtu != -1) {
+				*p++ = DHO_MAXMESSAGESIZE;
+				*p++ = 2;
+				sz = htons((uint16_t)mtu);
+				memcpy(p, &sz, 2);
+				p += 2;
+			}
 		}
 
 		if (ifo->userclass[0]) {
@@ -1160,6 +1238,13 @@ read_lease(struct interface *ifp)
 		else
 			logger(ifp->ctx, LOG_DEBUG,
 			    "%s: accepted reconfigure key", ifp->name);
+	} else if ((ifp->options->auth.options & DHCPCD_AUTH_SENDREQUIRE) ==
+	    DHCPCD_AUTH_SENDREQUIRE)
+	{
+		logger(ifp->ctx, LOG_ERR,
+		    "%s: authentication now required", ifp->name);
+		free(dhcp);
+		return NULL;
 	}
 
 	return dhcp;
@@ -1274,35 +1359,37 @@ dhcp_env(char **env, const char *prefix, const struct dhcp_message *dhcp,
 		/* Set some useful variables that we derive from the DHCP
 		 * message but are not necessarily in the options */
 		addr.s_addr = dhcp->yiaddr ? dhcp->yiaddr : dhcp->ciaddr;
-		setvar(ifp->ctx, &ep, prefix, "ip_address", inet_ntoa(addr));
+		addvar(ifp->ctx, &ep, prefix, "ip_address", inet_ntoa(addr));
 		if (get_option_addr(ifp->ctx, &net,
-		    dhcp, DHO_SUBNETMASK) == -1) {
+		    dhcp, DHO_SUBNETMASK) == -1)
+		{
 			net.s_addr = ipv4_getnetmask(addr.s_addr);
-			setvar(ifp->ctx, &ep, prefix,
+			addvar(ifp->ctx, &ep, prefix,
 			    "subnet_mask", inet_ntoa(net));
 		}
 		snprintf(cidr, sizeof(cidr), "%d", inet_ntocidr(net));
-		setvar(ifp->ctx, &ep, prefix, "subnet_cidr", cidr);
+		addvar(ifp->ctx, &ep, prefix, "subnet_cidr", cidr);
 		if (get_option_addr(ifp->ctx, &brd,
-		    dhcp, DHO_BROADCAST) == -1) {
+		    dhcp, DHO_BROADCAST) == -1)
+		{
 			brd.s_addr = addr.s_addr | ~net.s_addr;
-			setvar(ifp->ctx, &ep, prefix,
+			addvar(ifp->ctx, &ep, prefix,
 			    "broadcast_address", inet_ntoa(brd));
 		}
 		addr.s_addr = dhcp->yiaddr & net.s_addr;
-		setvar(ifp->ctx, &ep, prefix,
+		addvar(ifp->ctx, &ep, prefix,
 		    "network_number", inet_ntoa(addr));
 	}
 
 	if (*dhcp->bootfile && !(overl & 1)) {
 		print_string(safe, sizeof(safe), STRING,
 		    dhcp->bootfile, sizeof(dhcp->bootfile));
-		setvar(ifp->ctx, &ep, prefix, "filename", safe);
+		addvar(ifp->ctx, &ep, prefix, "filename", safe);
 	}
 	if (*dhcp->servername && !(overl & 2)) {
 		print_string(safe, sizeof(safe), STRING | DOMAIN,
 		    dhcp->servername, sizeof(dhcp->servername));
-		setvar(ifp->ctx, &ep, prefix, "server_name", safe);
+		addvar(ifp->ctx, &ep, prefix, "server_name", safe);
 	}
 
 	/* Zero our indexes */
@@ -1437,7 +1524,7 @@ dhcp_close(struct interface *ifp)
 		return;
 
 	if (state->raw_fd != -1) {
-		eloop_event_delete(ifp->ctx->eloop, state->raw_fd, 0);
+		eloop_event_delete(ifp->ctx->eloop, state->raw_fd);
 		close(state->raw_fd);
 		state->raw_fd = -1;
 	}
@@ -1457,19 +1544,8 @@ dhcp_openudp(struct interface *ifp)
 	char *p;
 #endif
 
-#ifdef SOCK_CLOEXEC
-	if ((s = socket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, IPPROTO_UDP)) == -1)
+	if ((s = xsocket(PF_INET, SOCK_DGRAM, IPPROTO_UDP, O_CLOEXEC)) == -1)
 		return -1;
-#else
-	if ((s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP)) == -1)
-		return -1;
-	if ((n = fcntl(s, F_GETFD, 0)) == -1 ||
-	    fcntl(s, F_SETFD, n | FD_CLOEXEC) == -1)
-	{
-		close(s);
-	        return -1;
-	}
-#endif
 
 	n = 1;
 	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &n, sizeof(n)) == -1)
@@ -1593,12 +1669,16 @@ send_message(struct interface *ifp, uint8_t type,
 	struct ipv4_addr *ia;
 #endif
 
-	if (!callback)
+	s = -1;
+	if (!callback) {
+		/* No carrier? Don't bother sending the packet. */
+		if (ifp->carrier == LINK_DOWN)
+			return;
 		logger(ifp->ctx, LOG_DEBUG, "%s: sending %s with xid 0x%x",
 		    ifp->name,
 		    ifo->options & DHCPCD_BOOTP ? "BOOTP" : get_dhcp_op(type),
 		    state->xid);
-	else {
+	} else {
 		if (state->interval == 0)
 			state->interval = 4;
 		else {
@@ -1610,6 +1690,10 @@ send_message(struct interface *ifp, uint8_t type,
 		tv.tv_nsec = (suseconds_t)arc4random_uniform(
 		    (DHCP_RAND_MAX - DHCP_RAND_MIN) * NSEC_PER_SEC);
 		timespecnorm(&tv);
+		/* No carrier? Don't bother sending the packet.
+		 * However, we do need to advance the timeout. */
+		if (ifp->carrier == LINK_DOWN)
+			goto fail;
 		logger(ifp->ctx, LOG_DEBUG,
 		    "%s: sending %s (xid 0x%x), next in %0.1f seconds",
 		    ifp->name,
@@ -1628,35 +1712,29 @@ send_message(struct interface *ifp, uint8_t type,
 	    ((ia = ipv4_iffindaddr(ifp, &state->addr, NULL)) &&
 	    !(ia->addr_flags & IN_IFF_NOTUSEABLE)) &&
 #endif
-	    (state->new->cookie == htonl(MAGIC_COOKIE) ||
-	    ifp->options->options & DHCPCD_INFORM))
+	    (state->lease.server.s_addr ||
+	    ifp->options->options & DHCPCD_INFORM) &&
+	    !IS_BOOTP(ifp, state->new))
 	{
 		s = dhcp_openudp(ifp);
-		if (s == -1 && errno != EADDRINUSE)
-			logger(ifp->ctx, LOG_ERR,
-			    "%s: dhcp_openudp: %m", ifp->name);
-	} else
-		s = -1;
-
-	/* If we couldn't open a UDP port for our IP address
-	 * then we cannot renew.
-	 * This could happen if our IP was pulled out from underneath us.
-	 * Also, we should not unicast from a BOOTP lease. */
-	if (s == -1 ||
-	    (!(ifo->options & DHCPCD_INFORM) &&
-	    IS_BOOTP(ifp, state->new)))
-	{
-		a = state->addr.s_addr;
-		state->addr.s_addr = INADDR_ANY;
+		if (s == -1) {
+			if (errno != EADDRINUSE)
+				logger(ifp->ctx, LOG_ERR,
+				    "%s: dhcp_openudp: %m", ifp->name);
+			/* We cannot renew */
+			a = state->addr.s_addr;
+			state->addr.s_addr = INADDR_ANY;
+		}
 	}
+
 	r = make_message(&dhcp, ifp, type);
+	if (a != INADDR_ANY)
+		state->addr.s_addr = a;
 	if (r == -1)
 		goto fail;
 	len = (size_t)r;
-	if (a)
-		state->addr.s_addr = a;
 	from.s_addr = dhcp->ciaddr;
-	if (from.s_addr)
+	if (s != -1 && from.s_addr != INADDR_ANY)
 		to.s_addr = state->lease.server.s_addr;
 	else
 		to.s_addr = INADDR_ANY;
@@ -1767,8 +1845,7 @@ dhcp_discover(void *arg)
 	if (ifo->fallback)
 		eloop_timeout_add_sec(ifp->ctx->eloop,
 		    ifo->reboot, dhcp_fallback, ifp);
-	else if (ifo->options & DHCPCD_IPV4LL &&
-	    !IN_LINKLOCAL(htonl(state->addr.s_addr)))
+	else if (ifo->options & DHCPCD_IPV4LL)
 		eloop_timeout_add_sec(ifp->ctx->eloop,
 		    ifo->reboot, ipv4ll_start, ifp);
 	if (ifo->options & DHCPCD_REQUEST)
@@ -1831,14 +1908,12 @@ dhcp_renew(void *arg)
 	send_renew(ifp);
 }
 
-#ifndef IN_IFF_TENTATIVE
 static void
 dhcp_arp_announced(struct arp_state *astate)
 {
 
 	arp_close(astate->iface);
 }
-#endif
 
 static void
 dhcp_rebind(void *arg)
@@ -1859,16 +1934,112 @@ dhcp_rebind(void *arg)
 	send_rebind(ifp);
 }
 
+static void
+dhcp_arp_probed(struct arp_state *astate)
+{
+	struct dhcp_state *state;
+	struct if_options *ifo;
+
+	state = D_STATE(astate->iface);
+	ifo = astate->iface->options;
+	if (state->arping_index < ifo->arping_len) {
+		/* We didn't find a profile for this
+		 * address or hwaddr, so move to the next
+		 * arping profile */
+		if (++state->arping_index < ifo->arping_len) {
+			astate->addr.s_addr =
+			    ifo->arping[state->arping_index - 1];
+			arp_probe(astate);
+		}
+		dhcpcd_startinterface(astate->iface);
+		return;
+	}
+
+	logger(astate->iface->ctx, LOG_DEBUG, "%s: DAD completed for %s",
+	    astate->iface->name, inet_ntoa(astate->addr));
+	dhcp_bind(astate->iface);
+	arp_announce(astate);
+
+	/* Stop IPv4LL now we have a working DHCP address */
+	ipv4ll_drop(astate->iface);
+}
+
+static void
+dhcp_arp_conflicted(struct arp_state *astate, const struct arp_msg *amsg)
+{
+	struct interface *ifp;
+	struct dhcp_state *state;
+	struct if_options *ifo;
+
+	ifp = astate->iface;
+	ifo = ifp->options;
+	state = D_STATE(ifp);
+	if (state->arping_index &&
+	    state->arping_index <= ifo->arping_len &&
+	    amsg &&
+	    (amsg->sip.s_addr == ifo->arping[state->arping_index - 1] ||
+	    (amsg->sip.s_addr == 0 &&
+	    amsg->tip.s_addr == ifo->arping[state->arping_index - 1])))
+	{
+		char buf[HWADDR_LEN * 3];
+
+		astate->failed.s_addr = ifo->arping[state->arping_index - 1];
+		arp_report_conflicted(astate, amsg);
+		hwaddr_ntoa(amsg->sha, ifp->hwlen, buf, sizeof(buf));
+		if (dhcpcd_selectprofile(ifp, buf) == -1 &&
+		    dhcpcd_selectprofile(ifp,
+		        inet_ntoa(astate->failed)) == -1)
+		{
+			/* We didn't find a profile for this
+			 * address or hwaddr, so move to the next
+			 * arping profile */
+			dhcp_arp_probed(astate);
+			return;
+		}
+		dhcp_close(ifp);
+		arp_free(astate);
+		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+		dhcpcd_startinterface(ifp);
+	}
+
+	/* RFC 2131 3.1.5, Client-server interaction
+	 * NULL amsg means IN_IFF_DUPLICATED */
+	if (amsg == NULL || (state->offer &&
+	    (amsg->sip.s_addr == state->offer->yiaddr ||
+	    (amsg->sip.s_addr == 0 &&
+	    amsg->tip.s_addr == state->offer->yiaddr))))
+	{
+#ifdef IN_IFF_DUPLICATED
+		struct ipv4_addr *ia;
+#endif
+
+		if (amsg)
+			astate->failed.s_addr = state->offer->yiaddr;
+		else
+			astate->failed = astate->addr;
+		arp_report_conflicted(astate, amsg);
+		unlink(state->leasefile);
+		if (!state->lease.frominfo)
+			dhcp_decline(ifp);
+#ifdef IN_IFF_DUPLICATED
+		ia = ipv4_iffindaddr(ifp, &astate->addr, NULL);
+		if (ia)
+			ipv4_deladdr(ifp, &ia->addr, &ia->net, 1);
+#endif
+		arp_free(astate);
+		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+		eloop_timeout_add_sec(ifp->ctx->eloop,
+		    DHCP_RAND_MAX, dhcp_discover, ifp);
+	}
+}
+
 void
-dhcp_bind(struct interface *ifp, struct arp_state *astate)
+dhcp_bind(struct interface *ifp)
 {
 	struct dhcp_state *state = D_STATE(ifp);
 	struct if_options *ifo = ifp->options;
 	struct dhcp_lease *lease = &state->lease;
-	uint8_t ipv4ll = 0;
 
-	if (state->state == DHS_BOUND)
-		goto applyaddr;
 	state->reason = NULL;
 	free(state->old);
 	state->old = state->new;
@@ -1881,12 +2052,6 @@ dhcp_bind(struct interface *ifp, struct arp_state *astate)
 		    inet_ntocidr(lease->net));
 		lease->leasetime = ~0U;
 		state->reason = "STATIC";
-	} else if (state->new->cookie != htonl(MAGIC_COOKIE)) {
-		logger(ifp->ctx, LOG_INFO, "%s: using IPv4LL address %s",
-		    ifp->name, inet_ntoa(lease->addr));
-		lease->leasetime = ~0U;
-		state->reason = "IPV4LL";
-		ipv4ll = 1;
 	} else if (ifo->options & DHCPCD_INFORM) {
 		if (ifo->req_addr.s_addr != 0)
 			lease->addr.s_addr = ifo->req_addr.s_addr;
@@ -1974,43 +2139,14 @@ dhcp_bind(struct interface *ifp, struct arp_state *astate)
 		    " seconds",
 		    ifp->name, lease->renewaltime, lease->rebindtime);
 	}
-	if (!(ifo->options & DHCPCD_STATIC) &&
-	    state->new->cookie != htonl(MAGIC_COOKIE))
-		state->state = DHS_IPV4LL_BOUND;
-	else
-		state->state = DHS_BOUND;
+	state->state = DHS_BOUND;
 	if (!state->lease.frominfo &&
 	    !(ifo->options & (DHCPCD_INFORM | DHCPCD_STATIC)))
 		if (write_lease(ifp, state->new) == -1)
 			logger(ifp->ctx, LOG_ERR,
 			    "%s: write_lease: %m", __func__);
 
-applyaddr:
 	ipv4_applyaddr(ifp);
-	if (ifo->options & DHCPCD_ARP &&
-	    !(ifp->ctx->options & DHCPCD_FORKED))
-	{
-#ifdef IN_IFF_TENTATIVE
-		if (astate)
-			arp_free_but(astate);
-		else if (!ipv4ll)
-			arp_close(ifp);
-#else
-		if (state->added) {
-			if (astate == NULL) {
-				astate = arp_new(ifp, &state->addr);
-				astate->announced_cb =
-				    dhcp_arp_announced;
-			}
-			if (astate) {
-				arp_announce(astate);
-				if (!ipv4ll)
-					arp_free_but(astate);
-			}
-		} else if (!ipv4ll)
-			arp_close(ifp);
-#endif
-	}
 }
 
 static void
@@ -2019,7 +2155,7 @@ dhcp_timeout(void *arg)
 	struct interface *ifp = arg;
 	struct dhcp_state *state = D_STATE(ifp);
 
-	dhcp_bind(ifp, NULL);
+	dhcp_bind(ifp);
 	state->interval = 0;
 	dhcp_discover(ifp);
 }
@@ -2034,6 +2170,7 @@ dhcp_message_new(const struct in_addr *addr, const struct in_addr *mask)
 	if (dhcp == NULL)
 		return NULL;
 	dhcp->yiaddr = addr->s_addr;
+	dhcp->cookie = htonl(MAGIC_COOKIE);
 	p = dhcp->options;
 	if (mask && mask->s_addr != INADDR_ANY) {
 		*p++ = DHO_SUBNETMASK;
@@ -2050,10 +2187,15 @@ dhcp_static(struct interface *ifp)
 {
 	struct if_options *ifo;
 	struct dhcp_state *state;
+	struct ipv4_addr *ia;
 
 	state = D_STATE(ifp);
 	ifo = ifp->options;
-	if (ifo->req_addr.s_addr == INADDR_ANY) {
+
+	ia = NULL;
+	if (ifo->req_addr.s_addr == INADDR_ANY &&
+	    (ia = ipv4_iffindaddr(ifp, NULL, NULL)) == NULL)
+	{
 		logger(ifp->ctx, LOG_INFO,
 		    "%s: waiting for 3rd party to "
 		    "configure IP address",
@@ -2062,10 +2204,12 @@ dhcp_static(struct interface *ifp)
 		script_runreason(ifp, state->reason);
 		return;
 	}
-	state->offer = dhcp_message_new(&ifo->req_addr, &ifo->req_mask);
+
+	state->offer = dhcp_message_new(ia ? &ia->addr : &ifo->req_addr,
+	    ia ? &ia->net : &ifo->req_mask);
 	if (state->offer) {
 		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-		dhcp_bind(ifp, NULL);
+		dhcp_bind(ifp);
 	}
 }
 
@@ -2101,7 +2245,7 @@ dhcp_inform(struct interface *ifp)
 			    dhcp_message_new(&ifo->req_addr, &ifo->req_mask);
 		if (state->offer) {
 			ifo->options |= DHCPCD_STATIC;
-			dhcp_bind(ifp, NULL);
+			dhcp_bind(ifp);
 			ifo->options &= ~DHCPCD_STATIC;
 		}
 	}
@@ -2170,8 +2314,7 @@ dhcp_reboot(struct interface *ifp)
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 
 	/* Need to add this before dhcp_expire and friends. */
-	if (!ifo->fallback && ifo->options & DHCPCD_IPV4LL &&
-	    !IN_LINKLOCAL(htonl(state->addr.s_addr)))
+	if (!ifo->fallback && ifo->options & DHCPCD_IPV4LL)
 		eloop_timeout_add_sec(ifp->ctx->eloop,
 		    ifo->reboot, ipv4ll_start, ifp);
 
@@ -2202,22 +2345,18 @@ dhcp_drop(struct interface *ifp, const char *reason)
 		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 		return;
 	}
-	/* Don't reset DHCP state if we have an IPv4LL address and link is up,
-	 * unless the interface is departing. */
-	if (state->state != DHS_IPV4LL_BOUND ||
-	    ifp->carrier != LINK_UP ||
-	    ifp->options->options & DHCPCD_DEPARTED)
-	{
-		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-		dhcp_auth_reset(&state->auth);
-		dhcp_close(ifp);
-	}
 
 	if (ifp->options->options & DHCPCD_RELEASE) {
+		/* Failure to send the release may cause this function to
+		 * re-enter so guard by setting the state. */
+		if (state->state == DHS_RELEASE)
+			return;
+		state->state = DHS_RELEASE;
+
 		unlink(state->leasefile);
 		if (ifp->carrier != LINK_DOWN &&
 		    state->new != NULL &&
-		    state->new->cookie == htonl(MAGIC_COOKIE))
+		    state->lease.server.s_addr != INADDR_ANY)
 		{
 			logger(ifp->ctx, LOG_INFO, "%s: releasing lease of %s",
 			    ifp->name, inet_ntoa(state->lease.addr));
@@ -2232,6 +2371,12 @@ dhcp_drop(struct interface *ifp, const char *reason)
 		}
 	}
 
+	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+	dhcp_auth_reset(&state->auth);
+	dhcp_close(ifp);
+
+	free(state->offer);
+	state->offer = NULL;
 	free(state->old);
 	state->old = state->new;
 	state->new = NULL;
@@ -2341,104 +2486,6 @@ whitelisted_ip(const struct if_options *ifo, in_addr_t addr)
 	return 0;
 }
 
-static void
-dhcp_arp_probed(struct arp_state *astate)
-{
-	struct dhcp_state *state;
-	struct if_options *ifo;
-
-	/* We didn't find a profile for this
-	 * address or hwaddr, so move to the next
-	 * arping profile */
-	state = D_STATE(astate->iface);
-	ifo = astate->iface->options;
-	if (state->arping_index < ifo->arping_len) {
-		if (++state->arping_index < ifo->arping_len) {
-			astate->addr.s_addr =
-			    ifo->arping[state->arping_index - 1];
-			arp_probe(astate);
-		}
-		dhcpcd_startinterface(astate->iface);
-		return;
-	}
-	dhcp_close(astate->iface);
-	eloop_timeout_delete(astate->iface->ctx->eloop, NULL, astate->iface);
-#ifdef IN_IFF_TENTATIVE
-	ipv4_finaliseaddr(astate->iface);
-	arp_close(astate->iface);
-#else
-	dhcp_bind(astate->iface, astate);
-#endif
-}
-
-static void
-dhcp_arp_conflicted(struct arp_state *astate, const struct arp_msg *amsg)
-{
-	struct dhcp_state *state;
-	struct if_options *ifo;
-
-	state = D_STATE(astate->iface);
-	ifo = astate->iface->options;
-	if (state->arping_index &&
-	    state->arping_index <= ifo->arping_len &&
-	    amsg &&
-	    (amsg->sip.s_addr == ifo->arping[state->arping_index - 1] ||
-	    (amsg->sip.s_addr == 0 &&
-	    amsg->tip.s_addr == ifo->arping[state->arping_index - 1])))
-	{
-		char buf[HWADDR_LEN * 3];
-
-		astate->failed.s_addr = ifo->arping[state->arping_index - 1];
-		arp_report_conflicted(astate, amsg);
-		hwaddr_ntoa(amsg->sha, astate->iface->hwlen, buf, sizeof(buf));
-		if (dhcpcd_selectprofile(astate->iface, buf) == -1 &&
-		    dhcpcd_selectprofile(astate->iface,
-		        inet_ntoa(astate->failed)) == -1)
-		{
-			/* We didn't find a profile for this
-			 * address or hwaddr, so move to the next
-			 * arping profile */
-			dhcp_arp_probed(astate);
-			return;
-		}
-		dhcp_close(astate->iface);
-		arp_close(astate->iface);
-		eloop_timeout_delete(astate->iface->ctx->eloop, NULL,
-		    astate->iface);
-		dhcpcd_startinterface(astate->iface);
-	}
-
-	/* RFC 2131 3.1.5, Client-server interaction
-	 * NULL amsg means IN_IFF_DUPLICATED */
-	if (amsg == NULL || (state->offer &&
-	    (amsg->sip.s_addr == state->offer->yiaddr ||
-	    (amsg->sip.s_addr == 0 &&
-	    amsg->tip.s_addr == state->offer->yiaddr))))
-	{
-#ifdef IN_IFF_DUPLICATED
-		struct ipv4_addr *ia;
-#endif
-
-		if (amsg)
-			astate->failed.s_addr = state->offer->yiaddr;
-		else
-			astate->failed = astate->addr;
-		arp_report_conflicted(astate, amsg);
-		unlink(state->leasefile);
-		if (!state->lease.frominfo)
-			dhcp_decline(astate->iface);
-#ifdef IN_IFF_DUPLICATED
-		ia = ipv4_iffindaddr(astate->iface, &astate->addr, NULL);
-		if (ia)
-			ipv4_deladdr(astate->iface, &ia->addr, &ia->net);
-#endif
-		eloop_timeout_delete(astate->iface->ctx->eloop, NULL,
-		    astate->iface);
-		eloop_timeout_add_sec(astate->iface->ctx->eloop,
-		    DHCP_RAND_MAX, dhcp_discover, astate->iface);
-	}
-}
-
 static void
 dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
     const struct in_addr *from)
@@ -2453,8 +2500,8 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 	unsigned int i;
 	size_t auth_len;
 	char *msg;
-	struct arp_state *astate;
 	struct ipv4_addr *ia;
+	struct arp_state *astate;
 
 	/* We may have found a BOOTP server */
 	if (get_option_uint8(ifp->ctx, &type, dhcp, DHO_MESSAGETYPE) == -1)
@@ -2486,12 +2533,15 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		else
 			logger(ifp->ctx, LOG_DEBUG,
 			    "%s: accepted reconfigure key", ifp->name);
-	} else if (ifo->auth.options & DHCPCD_AUTH_REQUIRE) {
-		log_dhcp1(LOG_ERR, "no authentication", ifp, dhcp, from, 0);
-		return;
-	} else if (ifo->auth.options & DHCPCD_AUTH_SEND)
+	} else if (ifo->auth.options & DHCPCD_AUTH_SEND) {
+		if (ifo->auth.options & DHCPCD_AUTH_REQUIRE) {
+			log_dhcp1(LOG_ERR, "no authentication",
+			    ifp, dhcp, from, 0);
+			return;
+		}
 		log_dhcp1(LOG_WARNING, "no authentication",
 		    ifp, dhcp, from, 0);
+	}
 
 	/* RFC 3203 */
 	if (type == DHCP_FORCERENEW) {
@@ -2505,7 +2555,8 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		if (auth == NULL) {
 			log_dhcp(LOG_ERR, "unauthenticated Force Renew",
 			    ifp, dhcp, from);
-			return;
+			if (ifo->auth.options & DHCPCD_AUTH_REQUIRE)
+				return;
 		}
 		if (state->state != DHS_BOUND && state->state != DHS_INFORM) {
 			log_dhcp(LOG_DEBUG, "not bound, ignoring Force Renew",
@@ -2617,8 +2668,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 			    "%s: message: %s", ifp->name, msg);
 			free(msg);
 		}
-		if ((state->state == DHS_DISCOVER ||
-		    state->state == DHS_IPV4LL_BOUND) &&
+		if (state->state == DHS_DISCOVER &&
 		    get_option_uint8(ifp->ctx, &tmp, dhcp,
 		    DHO_AUTOCONFIGURE) == 0)
 		{
@@ -2626,24 +2676,13 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 			case 0:
 				log_dhcp(LOG_WARNING, "IPv4LL disabled from",
 				    ifp, dhcp, from);
-				dhcp_drop(ifp, "EXPIRE");
+				ipv4ll_drop(ifp);
 				arp_close(ifp);
-				eloop_timeout_delete(ifp->ctx->eloop,
-				    NULL, ifp);
-				eloop_timeout_add_sec(ifp->ctx->eloop,
-				    DHCP_MAX, dhcp_discover,
-				    ifp);
 				break;
 			case 1:
 				log_dhcp(LOG_WARNING, "IPv4LL enabled from",
 				    ifp, dhcp, from);
-				eloop_timeout_delete(ifp->ctx->eloop,
-				    NULL, ifp);
-				if (IN_LINKLOCAL(htonl(state->addr.s_addr)))
-					eloop_timeout_add_sec(ifp->ctx->eloop,
-					    DHCP_MAX, dhcp_discover, ifp);
-				else
-					ipv4ll_start(ifp);
+				ipv4ll_start(ifp);
 				break;
 			default:
 				logger(ifp->ctx, LOG_ERR,
@@ -2651,6 +2690,9 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 				    ifp->name, tmp);
 				break;
 			}
+			eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
+			eloop_timeout_add_sec(ifp->ctx->eloop,
+			    DHCP_MAX, dhcp_discover, ifp);
 		}
 		return;
 	}
@@ -2672,7 +2714,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    ifp, dhcp, from);
 		if (type)
 			dhcp_decline(ifp);
-		ipv4_deladdr(ifp, &ia->addr, &ia->net);
+		ipv4_deladdr(ifp, &ia->addr, &ia->net, 0);
 		eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
 		eloop_timeout_add_sec(ifp->ctx->eloop,
 		    DHCP_RAND_MAX, dhcp_discover, ifp);
@@ -2680,9 +2722,7 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 	}
 #endif
 
-	if ((type == 0 || type == DHCP_OFFER) &&
-	    (state->state == DHS_DISCOVER || state->state == DHS_IPV4LL_BOUND))
-	{
+	if ((type == 0 || type == DHCP_OFFER) && state->state == DHS_DISCOVER) {
 		lease->frominfo = 0;
 		lease->addr.s_addr = dhcp->yiaddr;
 		lease->cookie = dhcp->cookie;
@@ -2738,7 +2778,6 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 		    ifo->options &= ~DHCPCD_STATIC;
 	}
 
-
 	/* No NAK, so reset the backoff
 	 * We don't reset on an OFFER message because the server could
 	 * potentially NAK the REQUEST. */
@@ -2754,35 +2793,44 @@ dhcp_handledhcp(struct interface *ifp, struct dhcp_message **dhcpp,
 
 	lease->frominfo = 0;
 	eloop_timeout_delete(ifp->ctx->eloop, NULL, ifp);
-	astate = NULL;
 
-#ifndef IN_IFF_TENTATIVE
-	if (ifo->options & DHCPCD_ARP
-	    && state->addr.s_addr != state->offer->yiaddr)
-#endif
-	{
-		addr.s_addr = state->offer->yiaddr;
-#ifndef IN_IFF_TENTATIVE
-		/* If the interface already has the address configured
-		 * then we can't ARP for duplicate detection. */
-		ia = ipv4_findaddr(ifp->ctx, &addr);
-		if (ia) {
-#endif
-			astate = arp_new(ifp, &addr);
-			if (astate) {
+	addr.s_addr = state->offer->yiaddr;
+	/* If the interface already has the address configured
+	 * then we can't ARP for duplicate detection. */
+	ia = ipv4_findaddr(ifp->ctx, &addr);
+
+#ifdef IN_IFF_TENTATIVE
+	if (ia == NULL || ia->addr_flags & IN_IFF_NOTUSEABLE) {
+		if ((astate = arp_new(ifp, &addr)) != NULL) {
+			astate->probed_cb = dhcp_arp_probed;
+			astate->conflicted_cb = dhcp_arp_conflicted;
+			astate->announced_cb = dhcp_arp_announced;
+		}
+		if (ia == NULL) {
+			struct dhcp_lease l;
+
+			get_lease(ifp->ctx, &l, state->offer);
+			/* Add the address now, let the kernel handle DAD. */
+			ipv4_addaddr(ifp, &l.addr, &l.net, &l.brd);
+		}
+		return;
+	}
+#else
+	if (ifo->options & DHCPCD_ARP) {
+		if (ia == NULL) {
+			if ((astate = arp_new(ifp, &addr)) != NULL) {
 				astate->probed_cb = dhcp_arp_probed;
 				astate->conflicted_cb = dhcp_arp_conflicted;
-#ifndef IN_IFF_TENTATIVE
+				astate->announced_cb = dhcp_arp_announced;
+				/* We need to handle DAD. */
 				arp_probe(astate);
-#endif
 			}
-#ifndef IN_IFF_TENTATIVE
 			return;
 		}
-#endif
 	}
+#endif
 
-	dhcp_bind(ifp, astate);
+	dhcp_bind(ifp);
 }
 
 static size_t
@@ -2951,7 +2999,7 @@ dhcp_handleudp(void *arg)
 	 * from the raw fd */
 	if (read(ctx->udp_fd, buffer, sizeof(buffer)) == -1) {
 		logger(ctx, LOG_ERR, "%s: %m", __func__);
-		eloop_event_delete(ctx->eloop, ctx->udp_fd, 0);
+		eloop_event_delete(ctx->eloop, ctx->udp_fd);
 		close(ctx->udp_fd);
 		ctx->udp_fd = -1;
 	}
@@ -2999,10 +3047,9 @@ dhcp_dump(struct interface *ifp)
 	ifp->if_data[IF_DATA_DHCP] = state = calloc(1, sizeof(*state));
 	if (state == NULL)
 		goto eexit;
-	state->raw_fd = state->arp_fd = -1;
-	TAILQ_INIT(&state->arp_states);
+	state->raw_fd = -1;
 	dhcp_set_leasefile(state->leasefile, sizeof(state->leasefile),
-	    AF_INET, ifp, "");
+	    AF_INET, ifp);
 	state->new = read_lease(ifp);
 	if (state->new == NULL) {
 		logger(ifp->ctx, LOG_ERR, "%s: %s: %m",
@@ -3029,7 +3076,6 @@ dhcp_free(struct interface *ifp)
 		free(state->old);
 		free(state->new);
 		free(state->offer);
-		free(state->buffer);
 		free(state->clientid);
 		free(state);
 		ifp->if_data[IF_DATA_DHCP] = NULL;
@@ -3046,7 +3092,7 @@ dhcp_free(struct interface *ifp)
 	}
 	if (ifp == NULL) {
 		if (ctx->udp_fd != -1) {
-			eloop_event_delete(ctx->eloop, ctx->udp_fd, 0);
+			eloop_event_delete(ctx->eloop, ctx->udp_fd);
 			close(ctx->udp_fd);
 			ctx->udp_fd = -1;
 		}
@@ -3073,8 +3119,7 @@ dhcp_init(struct interface *ifp)
 		if (state == NULL)
 			return -1;
 		/* 0 is a valid fd, so init to -1 */
-		state->raw_fd = state->arp_fd = -1;
-		TAILQ_INIT(&state->arp_states);
+		state->raw_fd = -1;
 
 		/* Now is a good time to find IPv4 routes */
 		if_initrt(ifp);
@@ -3084,7 +3129,7 @@ dhcp_init(struct interface *ifp)
 	state->reason = "PREINIT";
 	state->nakoff = 0;
 	dhcp_set_leasefile(state->leasefile, sizeof(state->leasefile),
-	    AF_INET, ifp, "");
+	    AF_INET, ifp);
 
 	ifo = ifp->options;
 	/* We need to drop the leasefile so that dhcp_start
@@ -3177,7 +3222,7 @@ dhcp_start1(void *arg)
 	}
 
 	state = D_STATE(ifp);
-	state->start_uptime = uptime();
+	clock_gettime(CLOCK_MONOTONIC, &state->started);
 	free(state->offer);
 	state->offer = NULL;
 
@@ -3214,7 +3259,7 @@ dhcp_start1(void *arg)
 	}
 	/* We don't want to read the old lease if we NAK an old test */
 	nolease = state->offer && ifp->ctx->options & DHCPCD_TEST;
-	if (!nolease) {
+	if (!nolease && ifo->options & DHCPCD_DHCP) {
 		state->offer = read_lease(ifp);
 		/* Check the saved lease matches the type we want */
 		if (state->offer) {
@@ -3305,13 +3350,8 @@ dhcp_start1(void *arg)
 	}
 
 	if (!(ifo->options & DHCPCD_DHCP)) {
-		if (ifo->options & DHCPCD_IPV4LL) {
-			if (state->offer && state->offer->cookie != 0) {
-				free(state->offer);
-				state->offer = NULL;
-			}
+		if (ifo->options & DHCPCD_IPV4LL)
 			ipv4ll_start(ifp);
-		}
 		return;
 	}
 
@@ -3330,6 +3370,13 @@ dhcp_start(struct interface *ifp)
 		return;
 
 	/* No point in delaying a static configuration */
+	if (ifp->options->options & DHCPCD_STATIC ||
+	    !(ifp->options->options & DHCPCD_INITIAL_DELAY))
+	{
+		dhcp_start1(ifp);
+		return;
+	}
+
 	tv.tv_sec = DHCP_MIN_DELAY;
 	tv.tv_nsec = (suseconds_t)arc4random_uniform(
 	    (DHCP_MAX_DELAY - DHCP_MIN_DELAY) * NSEC_PER_SEC);
@@ -3341,6 +3388,13 @@ dhcp_start(struct interface *ifp)
 	eloop_timeout_add_tv(ifp->ctx->eloop, &tv, dhcp_start1, ifp);
 }
 
+void
+dhcp_abort(struct interface *ifp)
+{
+
+	eloop_timeout_delete(ifp->ctx->eloop, dhcp_start1, ifp);
+}
+
 void
 dhcp_handleifa(int cmd, struct interface *ifp,
 	const struct in_addr *addr,
