@@ -94,12 +94,14 @@
 #define O_CONTROLGRP		O_BASE + 34
 #define O_SLAAC			O_BASE + 35
 #define O_GATEWAY		O_BASE + 36
-#define O_PFXDLGMIX		O_BASE + 37
+#define O_NOUP			O_BASE + 37
 #define O_IPV6RA_AUTOCONF	O_BASE + 38
 #define O_IPV6RA_NOAUTOCONF	O_BASE + 39
 #define O_REJECT		O_BASE + 40
 #define O_IPV6RA_ACCEPT_NOPUBLIC	O_BASE + 41
 #define O_BOOTP			O_BASE + 42
+#define O_DEFINEND		O_BASE + 43
+#define O_NODELAY		O_BASE + 44
 
 const struct option cf_options[] = {
 	{"background",      no_argument,       NULL, 'b'},
@@ -175,6 +177,7 @@ const struct option cf_options[] = {
 	{"dev",             required_argument, NULL, O_DEV},
 	{"nodev",           no_argument,       NULL, O_NODEV},
 	{"define",          required_argument, NULL, O_DEFINE},
+	{"definend",        required_argument, NULL, O_DEFINEND},
 	{"define6",         required_argument, NULL, O_DEFINE6},
 	{"embed",           required_argument, NULL, O_EMBED},
 	{"encap",           required_argument, NULL, O_ENCAP},
@@ -190,9 +193,10 @@ const struct option cf_options[] = {
 	{"controlgroup",    required_argument, NULL, O_CONTROLGRP},
 	{"slaac",           required_argument, NULL, O_SLAAC},
 	{"gateway",         no_argument,       NULL, O_GATEWAY},
-	{"ia_pd_mix",       no_argument,       NULL, O_PFXDLGMIX},
 	{"reject",          required_argument, NULL, O_REJECT},
 	{"bootp",           no_argument,       NULL, O_BOOTP},
+	{"nodelay",         no_argument,       NULL, O_NODELAY},
+	{"noup",            no_argument,       NULL, O_NOUP},
 	{NULL,              0,                 NULL, '\0'}
 };
 
@@ -467,7 +471,6 @@ parse_addr(struct dhcpcd_ctx *ctx,
     struct in_addr *addr, struct in_addr *net, const char *arg)
 {
 	char *p;
-	int i;
 
 	if (arg == NULL || *arg == '\0') {
 		if (addr != NULL)
@@ -477,10 +480,13 @@ parse_addr(struct dhcpcd_ctx *ctx,
 		return 0;
 	}
 	if ((p = strchr(arg, '/')) != NULL) {
+		int e;
+		intmax_t i;
+
 		*p++ = '\0';
-		if (net != NULL &&
-		    (sscanf(p, "%d", &i) != 1 ||
-			inet_cidrtoaddr(i, net) != 0))
+		i = strtoi(p, NULL, 10, 0, 32, &e);
+		if (e != 0 ||
+		    (net != NULL && inet_cidrtoaddr((int)i, net) != 0))
 		{
 			logger(ctx, LOG_ERR, "`%s' is not a valid CIDR", p);
 			return -1;
@@ -519,11 +525,22 @@ set_option_space(struct dhcpcd_ctx *ctx,
 {
 
 #if !defined(INET) && !defined(INET6)
-	/* Satisfy use */
-	ctx = NULL;
+	UNUSED(ctx);
 #endif
 
 #ifdef INET6
+	if (strncmp(arg, "nd_", strlen("nd_")) == 0) {
+		*d = ctx->nd_opts;
+		*dl = ctx->nd_opts_len;
+		*od = ifo->nd_override;
+		*odl = ifo->nd_override_len;
+		*request = ifo->requestmasknd;
+		*require = ifo->requiremasknd;
+		*no = ifo->nomasknd;
+		*reject = ifo->rejectmasknd;
+		return arg + strlen("nd_");
+	}
+
 	if (strncmp(arg, "dhcp6_", strlen("dhcp6_")) == 0) {
 		*d = ctx->dhcp6_opts;
 		*dl = ctx->dhcp6_opts_len;
@@ -633,7 +650,7 @@ parse_option(struct dhcpcd_ctx *ctx, const char *ifname, struct if_options *ifo,
 	int e, i, t;
 	long l;
 	unsigned long u;
-	char *p = NULL, *fp, *np, **nconf;
+	char *p = NULL, *bp, *fp, *np, **nconf;
 	ssize_t s;
 	struct in_addr addr, addr2;
 	in_addr_t *naddr;
@@ -1466,6 +1483,12 @@ err_sla:
 		dop = &ifo->dhcp_override;
 		dop_len = &ifo->dhcp_override_len;
 		/* FALLTHROUGH */
+	case O_DEFINEND:
+		if (dop == NULL) {
+			dop = &ifo->nd_override;
+			dop_len = &ifo->nd_override_len;
+		}
+		/* FALLTHROUGH */
 	case O_DEFINE6:
 		if (dop == NULL) {
 			dop = &ifo->dhcp6_override;
@@ -1535,13 +1558,18 @@ err_sla:
 		/* length */
 		if (np) {
 			*np++ = '\0';
+			bp = NULL; /* No bitflag */
 			l = (long)strtou(np, NULL, 0, 0, LONG_MAX, &e);
 			if (e) {
-				logger(ctx, LOG_ERR, "failed to convert length");
+				logger(ctx,LOG_ERR, "failed to convert length");
 				return -1;
 			}
-		} else
+		} else {
 			l = 0;
+			bp = strchr(arg, '='); /* bitflag assignment */
+			if (bp)
+				*bp++ = '\0';
+		}
 		t = 0;
 		if (strcasecmp(arg, "request") == 0) {
 			t |= REQUEST;
@@ -1590,6 +1618,8 @@ err_sla:
 			t |= STRING;
 		else if (strcasecmp(arg, "byte") == 0)
 			t |= UINT8;
+		else if (strcasecmp(arg, "bitflags") == 0)
+			t |= BITFLAG;
 		else if (strcasecmp(arg, "uint16") == 0)
 			t |= UINT16;
 		else if (strcasecmp(arg, "int16") == 0)
@@ -1605,7 +1635,7 @@ err_sla:
 		else if (strcasecmp(arg, "ascii") == 0)
 			t |= STRING | ASCII;
 		else if (strcasecmp(arg, "domain") == 0)
-			t |= STRING | DOMAIN | RFC3397;
+			t |= STRING | DOMAIN | RFC1035;
 		else if (strcasecmp(arg, "dname") == 0)
 			t |= STRING | DOMAIN;
 		else if (strcasecmp(arg, "binhex") == 0)
@@ -1632,11 +1662,16 @@ err_sla:
 			l = 0;
 		}
 		if (t & ARRAY && t & (STRING | BINHEX) &&
-		    !(t & (RFC3397 | DOMAIN)))
+		    !(t & (RFC1035 | DOMAIN)))
 		{
 			logger(ctx, LOG_WARNING, "ignoring array for strings");
 			t &= ~ARRAY;
 		}
+		if (t & BITFLAG) {
+			if (bp == NULL)
+				logger(ctx, LOG_WARNING,
+				    "missing bitflag assignment");
+		}
 		/* variable */
 		if (!fp) {
 			if (!(t & OPTION)) {
@@ -1650,10 +1685,16 @@ err_sla:
 			fp = strwhite(arg);
 			if (fp)
 				*fp++ = '\0';
-			np = strdup(arg);
-			if (np == NULL) {
-				logger(ctx, LOG_ERR, "%s: %m", __func__);
-				return -1;
+			if (strcasecmp(arg, "reserved")) {
+				np = strdup(arg);
+				if (np == NULL) {
+					logger(ctx, LOG_ERR,
+					    "%s: %m", __func__);
+					return -1;
+				}
+			} else {
+				np = NULL;
+				t |= RESERVED;
 			}
 		}
 		if (opt != O_EMBED) {
@@ -1688,11 +1729,25 @@ err_sla:
 		ndop->type = t;
 		ndop->len = (size_t)l;
 		ndop->var = np;
+		if (bp) {
+			dl = strlen(bp);
+			memcpy(ndop->bitflags, bp, dl);
+			memset(ndop->bitflags + dl, 0,
+			    sizeof(ndop->bitflags) - dl);
+		} else
+			memset(ndop->bitflags, 0, sizeof(ndop->bitflags));
 		/* Save the define for embed and encap options */
-		if (opt == O_DEFINE || opt == O_DEFINE6 || opt == O_VENDOPT)
+		switch (opt) {
+		case O_DEFINE:
+		case O_DEFINEND:
+		case O_DEFINE6:
+		case O_VENDOPT:
 			*ldop = ndop;
-		else if (opt == O_ENCAP)
+			break;
+		case O_ENCAP:
 			*edop = ndop;
+			break;
+		}
 		break;
 	case O_VENDCLASS:
 		fp = strwhite(arg);
@@ -1949,6 +2004,9 @@ err_sla:
 	case O_GATEWAY:
 		ifo->options |= DHCPCD_GATEWAY;
 		break;
+	case O_NOUP:
+		ifo->options &= ~DHCPCD_IF_UP;
+		break;
 	case O_SLAAC:
 		if (strcmp(arg, "private") == 0 ||
 		    strcmp(arg, "stableprivate") == 0 ||
@@ -1957,12 +2015,12 @@ err_sla:
 		else
 			ifo->options &= ~DHCPCD_SLAACPRIVATE;
 		break;
-	case O_PFXDLGMIX:
-		ifo->options |= DHCPCD_PFXDLGMIX;
-		break;
 	case O_BOOTP:
 		ifo->options |= DHCPCD_BOOTP;
 		break;
+	case O_NODELAY:
+		ifo->options &= ~DHCPCD_INITIAL_DELAY;
+		break;
 	default:
 		return 0;
 	}
@@ -2045,7 +2103,7 @@ read_config(struct dhcpcd_ctx *ctx,
 	char *line, *buf, *option, *p;
 	size_t buflen;
 	ssize_t vlen;
-	int skip = 0, have_profile = 0;
+	int skip, have_profile, new_block, had_block;
 #ifndef EMBEDDED_CONFIG
 	const char * const *e;
 	size_t ol;
@@ -2062,7 +2120,8 @@ read_config(struct dhcpcd_ctx *ctx,
 		logger(ctx, LOG_ERR, "%s: %m", __func__);
 		return NULL;
 	}
-	ifo->options |= DHCPCD_DAEMONISE | DHCPCD_LINK;
+	ifo->options |= DHCPCD_DAEMONISE | DHCPCD_LINK | DHCPCD_INITIAL_DELAY;
+	ifo->options |= DHCPCD_IF_UP;
 #ifdef PLUGIN_DEV
 	ifo->options |= DHCPCD_DEV;
 #endif
@@ -2100,6 +2159,14 @@ read_config(struct dhcpcd_ctx *ctx,
 			ifo->dhcp_override_len = INITDEFINES;
 #endif
 
+#if defined(INET6) && defined(INITDEFINENDS)
+		ifo->nd_override =
+		    calloc(INITDEFINENDS, sizeof(*ifo->nd_override));
+		if (ifo->nd_override == NULL)
+			logger(ctx, LOG_ERR, "%s: %m", __func__);
+		else
+			ifo->nd_override_len = INITDEFINENDS;
+#endif
 #if defined(INET6) && defined(INITDEFINE6S)
 		ifo->dhcp6_override =
 		    calloc(INITDEFINE6S, sizeof(*ifo->dhcp6_override));
@@ -2172,15 +2239,24 @@ read_config(struct dhcpcd_ctx *ctx,
 		ifo->dhcp_override_len = 0;
 
 #ifdef INET6
+		ctx->nd_opts = ifo->nd_override;
+		ctx->nd_opts_len = ifo->nd_override_len;
 		ctx->dhcp6_opts = ifo->dhcp6_override;
 		ctx->dhcp6_opts_len = ifo->dhcp6_override_len;
 #else
+		for (i = 0, opt = ifo->nd_override;
+		    i < ifo->nd_override_len;
+		    i++, opt++)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->nd_override);
 		for (i = 0, opt = ifo->dhcp6_override;
 		    i < ifo->dhcp6_override_len;
 		    i++, opt++)
 			free_dhcp_opt_embenc(opt);
 		free(ifo->dhcp6_override);
 #endif
+		ifo->nd_override = NULL;
+		ifo->nd_override_len = 0;
 		ifo->dhcp6_override = NULL;
 		ifo->dhcp6_override_len = 0;
 
@@ -2202,6 +2278,8 @@ read_config(struct dhcpcd_ctx *ctx,
 		ifo->mtime = sb.st_mtime;
 
 	ldop = edop = NULL;
+	skip = have_profile = new_block = 0;
+	had_block = ifname == NULL ? 1 : 0;
 	while ((line = get_line(&buf, &buflen, fp))) {
 		option = strsep(&line, " \t");
 		if (line)
@@ -2214,10 +2292,16 @@ read_config(struct dhcpcd_ctx *ctx,
 			    *(p - 1) != '\\')
 				*p-- = '\0';
 		}
+		if (skip == 0 && new_block) {
+			had_block = 1;
+			new_block = 0;
+			ifo->options &= ~DHCPCD_WAITOPTS;
+		}
 		/* Start of an interface block, skip if not ours */
 		if (strcmp(option, "interface") == 0) {
 			char **n;
 
+			new_block = 1;
 			if (ifname && line && strcmp(line, ifname) == 0)
 				skip = 0;
 			else
@@ -2244,6 +2328,7 @@ read_config(struct dhcpcd_ctx *ctx,
 		}
 		/* Start of an ssid block, skip if not ours */
 		if (strcmp(option, "ssid") == 0) {
+			new_block = 1;
 			if (ssid && line && strcmp(line, ssid) == 0)
 				skip = 0;
 			else
@@ -2252,6 +2337,7 @@ read_config(struct dhcpcd_ctx *ctx,
 		}
 		/* Start of a profile block, skip if not ours */
 		if (strcmp(option, "profile") == 0) {
+			new_block = 1;
 			if (profile && line && strcmp(line, profile) == 0) {
 				skip = 0;
 				have_profile = 1;
@@ -2276,6 +2362,8 @@ read_config(struct dhcpcd_ctx *ctx,
 		return NULL;
 	}
 
+	if (!had_block)
+		ifo->options &= ~DHCPCD_WAITOPTS;
 	finish_config(ifo);
 	return ifo;
 }
@@ -2285,18 +2373,27 @@ add_options(struct dhcpcd_ctx *ctx, const char *ifname,
     struct if_options *ifo, int argc, char **argv)
 {
 	int oi, opt, r;
+	unsigned long long wait_opts;
 
 	if (argc == 0)
 		return 1;
 
 	optind = 0;
 	r = 1;
+	/* Don't apply the command line wait options to each interface,
+	 * only use the dhcpcd.conf entry for that. */
+	if (ifname != NULL)
+		wait_opts = ifo->options & DHCPCD_WAITOPTS;
 	while ((opt = getopt_long(argc, argv, IF_OPTS, cf_options, &oi)) != -1)
 	{
 		r = parse_option(ctx, ifname, ifo, opt, optarg, NULL, NULL);
 		if (r != 1)
 			break;
 	}
+	if (ifname != NULL) {
+		ifo->options &= ~DHCPCD_WAITOPTS;
+		ifo->options |= wait_opts;
+	}
 
 	finish_config(ifo);
 	return r;
@@ -2334,6 +2431,11 @@ free_options(struct if_options *ifo)
 		    opt++, ifo->dhcp_override_len--)
 			free_dhcp_opt_embenc(opt);
 		free(ifo->dhcp_override);
+		for (opt = ifo->nd_override;
+		    ifo->nd_override_len > 0;
+		    opt++, ifo->nd_override_len--)
+			free_dhcp_opt_embenc(opt);
+		free(ifo->nd_override);
 		for (opt = ifo->dhcp6_override;
 		    ifo->dhcp6_override_len > 0;
 		    opt++, ifo->dhcp6_override_len--)
